From 9bc82199a2f027927ba0a04370e772b0c2db0600 Mon Sep 17 00:00:00 2001
From: Thomas Munro <thomas.munro@gmail.com>
Date: Thu, 21 Nov 2019 15:42:54 +1300
Subject: [PATCH] Improve batch number algorithm for large hash joins.

Our algorithm for choosing batch numbers turned out not to work
effectively for multi-billion key hash tables.  We would try to use
more hash bits than we have, and effectively concetrate all tuples
into a smaller number of batches than we intended.  While ideally
we should switch to wider hashes, for now, change the algorithm to
one that effectively gives up bits from the bucket number instead,
leaving some buckets empty.

Author: Thomas Munro
Reported-by: James Coleman
Discussion: https://postgr.es/m/16104-dc11ed911f1ab9df%40postgresql.org
---
 src/backend/executor/nodeHash.c | 29 ++++++-----------------------
 src/include/port/pg_bitutils.h  | 12 ++++++++++++
 2 files changed, 18 insertions(+), 23 deletions(-)

diff --git a/src/backend/executor/nodeHash.c b/src/backend/executor/nodeHash.c
index d6f4eda097..08b7c1dedf 100644
--- a/src/backend/executor/nodeHash.c
+++ b/src/backend/executor/nodeHash.c
@@ -37,6 +37,7 @@
 #include "miscadmin.h"
 #include "pgstat.h"
 #include "port/atomics.h"
+#include "port/pg_bitutils.h"
 #include "utils/dynahash.h"
 #include "utils/lsyscache.h"
 #include "utils/memutils.h"
@@ -1875,21 +1876,9 @@ ExecHashGetHashValue(HashJoinTable hashtable,
  * Note: on-the-fly increases of nbatch must not change the bucket number
  * for a given hash code (since we don't move tuples to different hash
  * chains), and must only cause the batch number to remain the same or
- * increase.  Our algorithm is
- *		bucketno = hashvalue MOD nbuckets
- *		batchno = (hashvalue DIV nbuckets) MOD nbatch
- * where nbuckets and nbatch are both expected to be powers of 2, so we can
- * do the computations by shifting and masking.  (This assumes that all hash
- * functions are good about randomizing all their output bits, else we are
- * likely to have very skewed bucket or batch occupancy.)
- *
- * nbuckets and log2_nbuckets may change while nbatch == 1 because of dynamic
- * bucket count growth.  Once we start batching, the value is fixed and does
- * not change over the course of the join (making it possible to compute batch
- * number the way we do here).
- *
- * nbatch is always a power of 2; we increase it only by doubling it.  This
- * effectively adds one more bit to the top of the batchno.
+ * increase.  To minimize the overlap of bits used for batch and bucket,
+ * we take them from opposite ends of the hash value, and reverse the
+ * batch number's bits so that we only add new high bits.
  */
 void
 ExecHashGetBucketAndBatch(HashJoinTable hashtable,
@@ -1901,16 +1890,10 @@ ExecHashGetBucketAndBatch(HashJoinTable hashtable,
 	uint32		nbatch = (uint32) hashtable->nbatch;
 
 	if (nbatch > 1)
-	{
-		/* we can do MOD by masking, DIV by shifting */
-		*bucketno = hashvalue & (nbuckets - 1);
-		*batchno = (hashvalue >> hashtable->log2_nbuckets) & (nbatch - 1);
-	}
+		*batchno = reverse_bits_u32(hashvalue) & (nbatch - 1);
 	else
-	{
-		*bucketno = hashvalue & (nbuckets - 1);
 		*batchno = 0;
-	}
+	*bucketno = hashvalue & (nbuckets - 1);
 }
 
 /*
diff --git a/src/include/port/pg_bitutils.h b/src/include/port/pg_bitutils.h
index 5197926696..210169739c 100644
--- a/src/include/port/pg_bitutils.h
+++ b/src/include/port/pg_bitutils.h
@@ -136,4 +136,16 @@ extern int	(*pg_popcount64) (uint64 word);
 /* Count the number of one-bits in a byte array */
 extern uint64 pg_popcount(const char *buf, int bytes);
 
+/*
+ * Reverse the bits in a uint32 value.  From figure 7-1 in Hacker's Delight.
+ */
+static inline uint32
+reverse_bits_u32(uint32 x)
+{
+	x = (x & 0x55555555) <<  1 | ((x >>  1) & 0x55555555);
+	x = (x & 0x33333333) <<  2 | ((x >>  2) & 0x33333333);
+	x = (x & 0x0F0F0F0F) <<  4 | ((x >>  4) & 0x0F0F0F0F);
+	return (x << 24) | ((x & 0xFF00) << 8) | ((x >> 8) & 0xFF00) | (x >> 24);
+}
+
 #endif							/* PG_BITUTILS_H */
-- 
2.23.0

